/**
  @file main.c
 */
/*
 * ============================================================================
 * History
 * =======
 * 2017-03-15 : Created
 *
 * (C) Copyright Bridgetek Pte Ltd
 * ============================================================================
 *
 * This source code ("the Software") is provided by Bridgetek Pte Ltd
 * ("Bridgetek") subject to the licence terms set out
 * http://www.ftdichip.com/FTSourceCodeLicenceTerms.htm ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of Bridgetek liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on Bridgetek's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, Bridgetek
 * has no liability in relation to those amendments.
 * ============================================================================
 */

/* INCLUDES ************************************************************************/

#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <ft900.h>
#include <ft900_dlog.h>
#include <ft900_uart_simple.h>
#include <ft900_eth.h>
#include <ft900_startup_dfu.h>

#include "..\libtelnet\libtelnet.h"

/* UART support for printf output. */
#include "tinyprintf.h"

#include "telnetd.h"
#include "net.h"

#include <lwip\apps\httpd.h>

#define BRIDGE_DEBUG
#ifdef BRIDGE_DEBUG
#define BRIDGE_DEBUG_PRINTF(...) do {tfp_printf(__VA_ARGS__);} while (0)
#else
#define BRIDGE_DEBUG_PRINTF(...)
#endif

/**
 @brief Link to datalogger area defined in crt0.S file.
 @details Must be passed to dlog library functions to initialise and use
        datalogger functions. We use the datalogger area for persistent
        configuration storage.
 */
extern __flash__ uint32_t __dlog_partition[];

/**
 @brief Pack structures to align on byte boundaries. Required for accurate layout
        of datalogger data structures.
 */
#define PACK __attribute__ ((__packed__))

/* CONSTANTS ***********************************************************************/

/**
 @brief Change this value in order to modify the size of the Tx and Rx ring buffers
  used to implement UART buffering. Make the ring buffer size 2^n for optimisation.
 */
#define RINGBUFFER_SIZE (1024)

/**
 @brief Page number in datalogger memory for network configuration.
 */
#define NET_CONFIG_PAGE 0

/**
 @brief Page number in datalogger memory for UART configuration.
 */
#define UART_CONFIG_PAGE 1

/**
 @brief Magic number to determine if structure stored in datalogger memory
        for network configuration is valid.
 */
#define NET_CONFIG_KEY 0x46383912

/**
 @brief Magic number to determine if structure stored in datalogger memory
        for UART configuration is valid.
 */
#define UART_CONFIG_KEY 0x49382913

/* GLOBAL VARIABLES ****************************************************************/

/* LOCAL VARIABLES *****************************************************************/

/**
 @brief Area reserved to store configuration data which will be updated in
        datalogger storage area when required.
 */
static uint8_t	flashbuf[256] __attribute__((aligned(4)));

/**
 @brief Millisecond counter
 @details Count-up timer to provide the elapsed time for network operations.
 */
static uint32_t milliseconds = 0;

/**
 @brief lwIP timer counter
 @details Count-down timer to provide the a periodic 'tick' for lwIP timer
 	 operations.
 */
static uint16_t lwip_timer = 0;

static uint16_t stats_timer = 0;

/**
 @brief telnetd handle
 @details Handle to the telnet daemon instance.
 */
static void *app_telnetd_handle;

/**
 @name UART ring buffer variables.
 @brief Variables required to implement an Rx and Tx ring buffer for the UART.
 */
//@{
static uint8_t uart0BufferIn_data[RINGBUFFER_SIZE];
static volatile uint16_t uart0BufferIn_wr_idx = 0;
static volatile uint16_t uart0BufferIn_rd_idx = 0;
static volatile uint16_t uart0BufferIn_avail = RINGBUFFER_SIZE;
static uint8_t uart0BufferOut_data[RINGBUFFER_SIZE];
static volatile uint16_t uart0BufferOut_wr_idx = 0;
static volatile uint16_t uart0BufferOut_rd_idx = 0;
static volatile uint16_t uart0BufferOut_avail = RINGBUFFER_SIZE;
//@}

/**
 @brief Structure to hold network settings.
 @details This is used to both keep current settings in use and
 	 to store persistently in Flash.
 */
struct net_config {
	uint32_t key;
	char hostname[64];
};

/**
 @brief Structure to hold UART settings.
 @details This is used to both keep current settings in use and
 	 to store persistently in Flash.
 */
struct uart_config {
	uint32_t key;
	int baud;
	uint8_t bits;
	uint8_t parity;
	uint8_t stop;
	uint8_t flow;
};

/**
 @brief Default network configuration.
 */
static struct net_config default_net_config = {
		NET_CONFIG_KEY,
		// Hostname.
		{'T', 'e', 'l', 'n', 'e', 't', 'B', 'r', 'i', 'd', 'g', 'e', 0,},
};

/**
 @brief Default UART configuration.
 */
static struct uart_config default_uart_config = {
		UART_CONFIG_KEY,
		19200,
		uart_data_bits_8,
		uart_parity_none,
		uart_stop_bits_1,
		0};

/**
 @brief Current network configuration.
 */
static struct net_config current_net_config;

/**
 @brief Current UART configuration.
 */
static struct uart_config current_uart_config;

/**
 @brief Watchdog keep alive flag.
 @details Set to zero to initiate a watchdog reset.
 */
int do_wdg_kick = 1;

/* MACROS **************************************************************************/

/* LOCAL FUNCTIONS / INLINES *******************************************************/

void flash_config_init(void);
void flash_config_write(void);
void flash_config_read(void);

size_t uart0Tx(uint8_t *data, size_t len);
void uart0Rx(void);

void uart0ISR(void);

/** @name tfp_putc
 *  @details Machine dependent putc function for tfp_printf (tinyprintf) library.
 *  @param p Parameters (machine dependent)
 *  @param c The character to write
 */
void tfp_putc(void* p, char c)
{
	uart_write((ft900_uart_regs_t*)p, (uint8_t)c);
}

/** @brief Returns the current millisecond country
 *  @returns A count of milliseconds
 */
uint32_t millis(void)
{
	return milliseconds;
}

void ISR_timer(void)
{
	if (timer_is_interrupted(timer_select_a))
	{
		milliseconds++;

		if (lwip_timer)
		{
			lwip_timer--;
		}

		if (stats_timer)
		{
			stats_timer--;
		}
	}
}

/**
 Transmit a buffer of data over UART0 asynchronously.

 @return The number of bytes written to the buffer
 */
size_t uart0Tx(uint8_t *buffer, size_t tot_len)
{
	signed int start;
	size_t avail;
	size_t len;
	size_t copied = 0;
	uint8_t c;

	do
	{
		// Check the data received from the telnet library does
		// not overflow the ring buffer.
		if (tot_len > RINGBUFFER_SIZE / 2)
		{
			// Fill half ring buffer only to allow sending to start
			// immediately.
			len = RINGBUFFER_SIZE / 2;
		}
		else
		{
			// Send all now.
			len = tot_len;
		}

		// Decrement by number of bytes transmitted (or scheduled to transmit).
		tot_len -= len;

		do
		{
			// Disable transmit interrupts while we check the available
			// space in the transmit buffer.
			uart_disable_interrupt(UART0, uart_interrupt_tx);
			// Check there is space there to receive the data from the producer.
			avail = uart0BufferOut_avail;
			uart_enable_interrupt(UART0, uart_interrupt_tx);

			// Sufficient space is available.
			if (avail >= len)
			{
				break;
			}

			// Short delay to allow ISR to transmit.
			delayus(10);
		} while (1);

		// Disable transmit interrupts while we fill up the transmit
		// buffer.
		uart_disable_interrupt(UART0, uart_interrupt_tx);

		// Determine if we need to start a transmission or there is one
		// already in action.
		if (len)
		{
			start = uart0BufferOut_wr_idx - uart0BufferOut_rd_idx;
		}
		else
		{
			start = 0;
		}

		// Copy in as much data from the producer as there is space for
		// in the out buffer.
		while (len--)
		{
			// Check there is space there to receive the data from the producer.
			if (uart0BufferOut_avail > 0)
			{
				// Reduce the number of available bytes.
				uart0BufferOut_avail--;

				// Add byte of data to the buffer.
				uart0BufferOut_data[uart0BufferOut_wr_idx] = *buffer++;

				// Increment the pointer and wrap around.
				uart0BufferOut_wr_idx++;
				uart0BufferOut_wr_idx &= (RINGBUFFER_SIZE - 1);

				// Tally up number of bytes actually transmitted.
				copied++;
			}
		}

		// Start new a transmission if nothing is being transmitted.
		if (start == 0)
		{
			// Get data from the out buffer.
			c = uart0BufferOut_data[uart0BufferOut_rd_idx];

			// Increment the consumer pointer and wrap around.
			uart0BufferOut_rd_idx++;
			uart0BufferOut_rd_idx &= (RINGBUFFER_SIZE - 1);

			// Byte has been sent, update available bytes value.
			uart0BufferOut_avail++;

			// Send data to the UART. This will result in a TX interrupt when
			// it completes.
			uart_write(UART0, c);
		}
		uart_enable_interrupt(UART0, uart_interrupt_tx);


	} while (tot_len);

	return copied;
}

/**
 Receive a number of bytes from UART0

 @return The number of bytes read from UART0
 */
void uart0Rx(void)
{
	size_t waiting = 0;
	uint8_t *pbuf;
	size_t len = 0;

	// Disable interrupts while we check the receive buffer.
	uart_disable_interrupt(UART0, uart_interrupt_rx);

	// Work out how many bytes are waiting to be received.
	waiting = RINGBUFFER_SIZE - uart0BufferIn_avail;
	if (waiting > 0)
	{
		// Get the start point of the data buffer.
		pbuf = &uart0BufferIn_data[uart0BufferIn_rd_idx];
		// Work out how many bytes until the end of the buffer.
		len = RINGBUFFER_SIZE - uart0BufferIn_rd_idx;
		if (waiting < len)
		{
			len = waiting;
		}
	}
	uart_enable_interrupt(UART0, uart_interrupt_rx);

	// Write any data to the telnet interface.
	if (len)
	{
		if (app_telnetd_handle != NULL)
		{
			telnetd_send(pbuf, len, app_telnetd_handle);
		}

		// Disable receive interrupts while we update the
		// buffer pointers.
		uart_disable_interrupt(UART0, uart_interrupt_rx);
		uart0BufferIn_avail += len;
		uart0BufferIn_rd_idx += len;
		uart0BufferIn_rd_idx &= (RINGBUFFER_SIZE - 1);
		uart_enable_interrupt(UART0, uart_interrupt_rx);
	}
}

/**
 The Interrupt which handles asynchronous transmission and reception
 of data into the ring buffer
 */
void uart0ISR(void)
{
	static uint8_t c;

	// Transmit interrupt...
	if (uart_is_interrupted(UART0, uart_interrupt_tx))
	{
		// Check to see how much data we have to transmit...
		//avail = uart0BufferOut_rd_idx - uart0BufferOut_wr_idx;

		if (uart0BufferOut_avail < RINGBUFFER_SIZE)
		{
			uart0BufferOut_avail++;

			// Copy out the byte to be transmitted so that the uart_write is
			// the last thing we do...
			c = uart0BufferOut_data[uart0BufferOut_rd_idx];

			// Increment the pointer and wrap around.
			uart0BufferOut_rd_idx++;
			uart0BufferOut_rd_idx &= (RINGBUFFER_SIZE - 1);

			// Write out a new byte, the following Transmit interrupt should handle
			// the remaning bytes...
			uart_write(UART0, c);
		}
	}

	// Receive interrupt...
	if (uart_is_interrupted(UART0, uart_interrupt_rx))
	{
		// Read a byte into the Ring Buffer...
		uart_read(UART0, &c);

		// Do not overwrite data already in the buffer.
		if (uart0BufferIn_avail > 0)
		{
			uart0BufferIn_avail--;

			uart0BufferIn_data[uart0BufferIn_wr_idx] = c;

			// Increment the pointer and wrap around.
			uart0BufferIn_wr_idx++;
			uart0BufferIn_wr_idx &= (RINGBUFFER_SIZE - 1);
		}
	}
}

void uart_config(void)
{
	uint16_t divisor;
	uint8_t prescaler;
	int32_t baud_deviation;

	// Have UART settings at this stage.
	baud_deviation = uart_calculate_baud(current_uart_config.baud, 4,
			sys_get_cpu_clock(), &divisor, &prescaler);

	BRIDGE_DEBUG_PRINTF("Baud requested %d (deviation %ld)\n", current_uart_config.baud, baud_deviation);

	if (((baud_deviation * 100)/current_uart_config.baud) > 3)
	{
		BRIDGE_DEBUG_PRINTF("Unobtainable baud rate using default UART configuration\n");
		prescaler = 1;
		divisor = UART_DIVIDER_19200_BAUD;
	}

	// Open the UART using the coding required.
	uart_open(UART0, prescaler, divisor, current_uart_config.bits,
			current_uart_config.parity, current_uart_config.stop);
}

// Callback of type telnetd_handle_cb.
void app_telnetd_handle_cb(void *handle)
{
	BRIDGE_DEBUG_PRINTF("Telnet Library Handle = %p\n", handle);
	app_telnetd_handle = handle;
}

// Callback to receive data from telnet library.
void app_telnetd_receive_cb(uint8_t *buf, size_t len, void *handle)
{
	if (handle == app_telnetd_handle)
	{
		uart0Tx(buf, len);
	}
}

void app_telnetd_error_cb(int err, void *handle)
{
	BRIDGE_DEBUG_PRINTF("Telnet Library Error %d (Handle = %p)\n", err, handle);
}

// HTTPD Server Side Include Handler
#define FOREACH_SSI(MAKE_SSI) \
		MAKE_SSI(mac)   \
		MAKE_SSI(hn)   \
		MAKE_SSI(ip)  \
		MAKE_SSI(gw)   \
		MAKE_SSI(nm)  \
		MAKE_SSI(dhcp)  \
		MAKE_SSI(rh)  \
		MAKE_SSI(baud)  \
		MAKE_SSI(bi)  \
		MAKE_SSI(pa)  \
		MAKE_SSI(sb)  \
		MAKE_SSI(fc)  \
		MAKE_SSI(total)

#define GENERATE_SSI_ENUM(ENUM) ssiTag_##ENUM,
#define GENERATE_SSI_STRING(STRING) #STRING,

enum SSI_ENUM {
	FOREACH_SSI(GENERATE_SSI_ENUM)
};

static const char *ssiTags[] = {
		FOREACH_SSI(GENERATE_SSI_STRING)
};

u16_t app_httpd_ssi_cb(int iIndex, char *pcInsert, int iInsertLen)
{
	int len = 0;
	uint8_t *pval;
	ip_addr_t ia;
	char rstring[32];
	char *enstring = "";
	char *dstring;
	const char *nstring;
	struct eeprom_net_config getval;

	net_get_eeprom(&getval);

	if (getval.dhcp == 0)
	{
		enstring = "disabled";
	}

	if (iIndex < ssiTag_total)
	{
		nstring = ssiTags[iIndex];

		switch (iIndex)
		{
		case ssiTag_mac:
			pval = net_get_mac(rstring);
			sprintf(rstring, "%02x:%02x:%02x:%02x:%02x:%02x",
					pval[0], pval[1], pval[2], pval[3], pval[4], pval[5]);
			len = sprintf(pcInsert, " <input type=\"text\" name=\"%s\" maxlength=\"18\" value=\"%s\" disabled>", nstring, rstring);
			break;
		case ssiTag_hn:
			len = sprintf(pcInsert, "<input type=\"text\" name=\"%s\" value=\"%s\" maxlength=\"64\">", nstring, current_net_config.hostname);
			break;
		case ssiTag_ip:
			ia = getval.ip;
			dstring = ipaddr_ntoa(&ia);
			len = sprintf(pcInsert, " <input type=\"text\" name=\"%s\" value=\"%s\" maxlength=\"16\" %s>", nstring, dstring, enstring);
			break;
		case ssiTag_gw:
			ia = getval.gw;
			dstring = ipaddr_ntoa(&ia);
			len = sprintf(pcInsert, " <input type=\"text\" name=\"%s\" value=\"%s\" maxlength=\"16\" %s>", nstring, dstring, enstring);
			break;
		case ssiTag_nm:
			ia = getval.mask;
			dstring = ipaddr_ntoa(&ia);
			len = sprintf(pcInsert, " <input type=\"text\" name=\"%s\" value=\"%s\" maxlength=\"16\" %s>", nstring, dstring, enstring);
			break;
		case ssiTag_dhcp:
			if (getval.dhcp)
			{
				dstring = "checked";
			}
			len = sprintf(pcInsert, " <input type=\"checkbox\" name=\"%s\" %s>", nstring, dstring);
			break;
		case ssiTag_rh:
			if (app_telnetd_handle)
			{
				ia = telnetd_remote_ip(app_telnetd_handle);
				dstring = ipaddr_ntoa(&ia);
			}
			else
			{
				dstring = "No connection";
			}
			len = strlen(dstring);
			strcpy(pcInsert, dstring);
			break;
		case ssiTag_baud:
			len = sprintf(pcInsert, "<input type=\"text\" name=\"%s\" value=\"%d\" maxlength=\"16\">", nstring, current_uart_config.baud);
			break;
		case ssiTag_bi:
			len = sprintf(pcInsert, "<input type=\"text\" name=\"%s\" value=\"%d\" min=\"5\" max=\"8\">", nstring, current_uart_config.bits);
			break;
		case ssiTag_pa:
			len = sprintf(pcInsert, "<input type=\"text\" name=\"%s\" value=\"%c\" maxlength=\"1\">", nstring, current_uart_config.parity?'Y':'N');
			break;
		case ssiTag_sb:
			len = sprintf(pcInsert, "<input type=\"text\" name=\"%s\" value=\"%d\" min=\"1\" max=\"2\">", nstring, current_uart_config.stop + 1);
			break;
		case ssiTag_fc:
			len = sprintf(pcInsert, "<input type=\"text\" name=\"%s\" value=\"none\" disabled>", nstring);
			break;
		}
	}

	return len;
}

// HTTPD CGI Handlers

const char *app_httpd_cgi_net(int iIndex, int iNumParams, char *pcParam[],
		char *pcValue[])
{
	int i;
	int mac[6];
	int val;
	struct eeprom_net_config setval;

	// Checkboxes not always sent if unchecked.
	net_get_eeprom(&setval);
	setval.dhcp = 0;

	for(i = 0; i < iNumParams; i++)
	{
		if(!strcmp(pcParam[i], ssiTags[ssiTag_mac]))
		{
			if (sscanf(pcValue[i], "%x:%x:%x:%x:%x:%x",
			    &mac[0], &mac[1], &mac[2],
			    &mac[3], &mac[4], &mac[5]) == 6)
			{
			    for (val = 0; val < 6; val++)
			    {
			        //current_net_config.macaddr[val] = (uint8_t) mac[val];
			    }
			}
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_hn]))
		{
			strcpy(current_net_config.hostname, pcValue[i]);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_dhcp]))
		{
			setval.dhcp = 1;
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_ip]))
		{
			ipaddr_aton(pcValue[i], &setval.ip);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_gw]))
		{
			ipaddr_aton(pcValue[i], &setval.gw);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_nm]))
		{
			ipaddr_aton(pcValue[i], &setval.mask);
		}
	}

	BRIDGE_DEBUG_PRINTF("Net Config:\n");
	//BRIDGE_DEBUG_PRINTF("  MAC = %02x:%02x:%02x:%02x:%02x:%02x\n",
		//	current_net_config.macaddr[0], current_net_config.macaddr[1],
		//	current_net_config.macaddr[2], current_net_config.macaddr[3],
		//	current_net_config.macaddr[4], current_net_config.macaddr[5]);
	BRIDGE_DEBUG_PRINTF("  Hostname = \"%s\"\n", current_net_config.hostname);
	BRIDGE_DEBUG_PRINTF("  DHCP = %c\n", setval.dhcp?'Y':'N');
	BRIDGE_DEBUG_PRINTF("  ipaddr = %s\n", ipaddr_ntoa(&setval.ip));
	BRIDGE_DEBUG_PRINTF("  gwaddr = %s\n", ipaddr_ntoa(&setval.gw));
	BRIDGE_DEBUG_PRINTF("  netmask = %s\n", ipaddr_ntoa(&setval.mask));

	net_update_eeprom(&setval);
	flash_config_write();

	// Force a reboot through the watchdog timer.
	do_wdg_kick = 0;

	return "/restart.shtm";
}

const char *app_httpd_cgi_uart(int iIndex, int iNumParams, char *pcParam[],
		char *pcValue[])
{
	int i;
	int val;

	for(i = 0; i < iNumParams; i++)
	{
		if(!strcmp(pcParam[i], ssiTags[ssiTag_baud]))
		{
			current_uart_config.baud = atoi(pcValue[i]);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_bi]))
		{
			val = atoi(pcValue[i]);
			switch (val)
			{
			case 5: current_uart_config.bits = uart_data_bits_5; break;
			case 6: current_uart_config.bits = uart_data_bits_6; break;
			case 7: current_uart_config.bits = uart_data_bits_7; break;
			case 8: current_uart_config.bits = uart_data_bits_8; break;
			}
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_pa]))
		{
			switch (*pcValue[i])
			{
			case 'N': case 'n': current_uart_config.parity = uart_parity_none; break;
			case 'O': case 'o': current_uart_config.parity = uart_parity_odd; break;
			case 'E': case 'e': current_uart_config.parity = uart_parity_even; break;
			}
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_sb]))
		{
			val = atoi(pcValue[i]);
			switch (val)
			{
			case 1: current_uart_config.stop = uart_stop_bits_1; break;
			case 2: current_uart_config.stop = uart_stop_bits_2; break;
			}
		}
	}

	BRIDGE_DEBUG_PRINTF("UART Config:\n");
	BRIDGE_DEBUG_PRINTF("  Baud = %d\n", current_uart_config.baud);
	BRIDGE_DEBUG_PRINTF("  Bits = %d\n", current_uart_config.bits);
	BRIDGE_DEBUG_PRINTF("  Stop = %d\n", current_uart_config.stop);
	BRIDGE_DEBUG_PRINTF("  Parity = %d\n", current_uart_config.parity);

	flash_config_write();

	uart_config();

	return "/update.html";
}

const tCGI cgiHandler[] = {
		{"/net", app_httpd_cgi_net},
		{"/uart", app_httpd_cgi_uart},
};

void flash_config_init(void)
{
	int	pgsz;
	int	i;

	if (dlog_init(__dlog_partition, &pgsz, &i) < 0)
	{
		// Project settings incorrect. Require dlog support with modified
		// linker script and crt0.S file.
		// See AN_398 for examples.
		return;
	}
}

void flash_config_write(void)
{
	dlog_erase();
	memset(flashbuf, 0xff, sizeof(flashbuf));
	memcpy(flashbuf, &current_net_config, sizeof(struct net_config));
	dlog_prog(NET_CONFIG_PAGE, (uint32_t *)flashbuf);
	memset(flashbuf, 0xff, sizeof(flashbuf));
	memcpy(flashbuf, &current_uart_config, sizeof(struct uart_config));
	dlog_prog(UART_CONFIG_PAGE, (uint32_t *)flashbuf);
}

void flash_config_read(void)
{
	dlog_read(NET_CONFIG_PAGE, (uint32_t *)flashbuf);
	memcpy(&current_net_config, flashbuf, sizeof(struct net_config));
	dlog_read(UART_CONFIG_PAGE, (uint32_t *)flashbuf);
	memcpy(&current_uart_config, flashbuf, sizeof(struct uart_config));
}

void bridge(void)
{
	ip_addr_t ip_addr = {0};
	struct telnetd_setup telnetdconfig;
	uint8_t flash_init = 0;

	flash_config_init();
	flash_config_read();

	// Implement defaults if required.
	if (current_net_config.key != NET_CONFIG_KEY)
	{
		BRIDGE_DEBUG_PRINTF("Using default network configuration\n");
		// Setup using default network configuration.
		memcpy(&current_net_config, &default_net_config, sizeof(struct net_config));
		flash_init = 1;
	}
	if (current_uart_config.key != UART_CONFIG_KEY)
	{
		BRIDGE_DEBUG_PRINTF("Using default UART configuration\n");
		// Setup using default UART configuration.
		memcpy(&current_uart_config, &default_uart_config, sizeof(struct uart_config));
		flash_init = 1;
	}
	if (flash_init != 0)
	{
		flash_config_write();
	}

	uart_config();

	telnetdconfig.hcb = app_telnetd_handle_cb;
	telnetdconfig.rcb = app_telnetd_receive_cb;
	telnetdconfig.ecb = app_telnetd_error_cb;

	wdt_kick();

	// Open the lwIP library and existing EEPROM settings.
	if (net_init(ip_addr, ip_addr, ip_addr, 0, current_net_config.hostname) == ERR_OK)
	{
		int started = 0;

		while (1)
		{
			// Read in a packet of data from the UART and
			// send it to the telnet daemon.
			uart0Rx();

			if (net_is_link_up())
			{
				net_tick();

				if (!started)
				{
					started = net_is_up();
					if (started)
					{
						httpd_init();
						http_set_ssi_handler(app_httpd_ssi_cb, ssiTags, ssiTag_total);
						http_set_cgi_handlers(cgiHandler, 2);
						telnetd_init(&telnetdconfig);
					}
				}
			}

			if (do_wdg_kick)
			{
				wdt_kick();
			}

			if (stats_timer == 0)
			{
				//stats_display();
				//netif_stats_display_ft900();
				stats_timer = 10000;
			}

			if (lwip_timer == 0)
			{
				// Check for ethernet disconnection.
				if (net_is_link_up())
				{
					if (!ethernet_is_link_up())
					{
						net_set_link_down();
						BRIDGE_DEBUG_PRINTF("Ethernet disconnected.\n");
					}
				}
				else
				{
					if (ethernet_is_link_up())
					{
						net_set_link_up();
						BRIDGE_DEBUG_PRINTF("Ethernet connected.\n");
					}
				}

				// Check lwIP timeouts every 250 ms.
				sys_check_timeouts();
				lwip_timer = 250;
			}
		}
	}
	// Disconnected from host.

	/* Set the I2C Master pins back to channel 0 */
	sys_i2c_swop(0);
}

/* FUNCTIONS ***********************************************************************/

int main(void)
{
	sys_reset_all();
    sys_disable(sys_device_i2c_master);

	/* Enable the UART Device... */
	sys_enable(sys_device_uart0);
	/* Make GPIO48 function as UART0_TXD and GPIO49 function as UART0_RXD... */
	gpio_function(48, pad_uart0_txd); /* UART0 TXD */
	gpio_function(49, pad_uart0_rxd); /* UART0 RXD */

	/* Enable tfp_printf() functionality... */
	init_printf(UART0, tfp_putc);

#ifdef BRIDGE_DEBUG
	// Open the UART using the coding required.
	uart_open(UART0, 1, UART_DIVIDER_19200_BAUD, uart_data_bits_8, uart_parity_none, uart_stop_bits_1);

	/* Print out a welcome message... */
	BRIDGE_DEBUG_PRINTF("\x1B[2J" /* ANSI/VT100 - Clear the Screen */
			"\x1B[H"  /* ANSI/VT100 - Move Cursor to Home */
	);

	BRIDGE_DEBUG_PRINTF("(C) Copyright 2016, Bridgetek Pte Ltd. \r\n");
	BRIDGE_DEBUG_PRINTF("--------------------------------------------------------------------- \r\n");
	BRIDGE_DEBUG_PRINTF("Welcome to the Telnet Bridge ... \r\n");
	BRIDGE_DEBUG_PRINTF("\r\n");
	BRIDGE_DEBUG_PRINTF("Emulate a Telnet device connected to the Ethernet.\r\n");
	BRIDGE_DEBUG_PRINTF("--------------------------------------------------------------------- \r\n");
#endif // BRIDGE_DEBUG

	sys_enable(sys_device_timer_wdt);

	/* Enable the UART to fire interrupts when receiving data... */
	uart_enable_interrupt(UART0, uart_interrupt_rx);
	/* Enable the UART to fire interrupts when transmitting data... */
	uart_enable_interrupt(UART0, uart_interrupt_tx);
	/* Attach the interrupt so it can be called... */
	interrupt_attach(interrupt_uart0, (uint8_t) interrupt_uart0, uart0ISR);
	/* Enable interrupts to be fired... */
	uart_enable_interrupts_globally(UART0);

	/* Timer A = 1ms */
	timer_prescaler(1000);
	timer_init(timer_select_a, 100, timer_direction_down, timer_prescaler_select_on, timer_mode_continuous);
	timer_enable_interrupt(timer_select_a);
	timer_start(timer_select_a);

	interrupt_attach(interrupt_timers, (int8_t)interrupt_timers, ISR_timer);

	// Start watchdog timer to restart the device after approx 1.34 seconds.
	wdt_init(wdt_counter_1G_clocks);

    i2cm_init(I2CM_NORMAL_SPEED, 10000);

	sys_enable(sys_device_ethernet);

    /* Set up I2C */
    sys_enable(sys_device_i2c_master);

    /* Setup I2C channel 0 pins */
    /* Use sys_i2c_swop(0) to activate. */
    gpio_function(44, pad_i2c0_scl); /* I2C0_SCL */
    gpio_function(45, pad_i2c0_sda); /* I2C0_SDA */

    /* Setup I2C channel 1 pins for EEPROM */
    /* Use sys_i2c_swop(1) to activate. */
    gpio_function(46, pad_i2c1_scl); /* I2C1_SCL */
    gpio_function(47, pad_i2c1_sda); /* I2C1_SDA */

	interrupt_enable_globally();

	bridge();

	interrupt_disable_globally();

	// Wait forever...
	for (;;);

	return 0;
}
