/**
  @file telnetd.c
  @brief Abstraction file between lwIP and libtelnet.
  @details Based on ECHO example code in lwIP and example code in libtelnet.
 */
/*
 * ============================================================================
 * History
 * =======
 * 2017-03-15 : Created
 *
 * (C) Copyright Bridgetek Pte Ltd
 * ============================================================================
 *
 * This source code ("the Software") is provided by Bridgetek Pte Ltd
 * ("Bridgetek") subject to the licence terms set out
 * http://www.ftdichip.com/FTSourceCodeLicenceTerms.htm ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of Bridgetek liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on Bridgetek's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, Bridgetek
 * has no liability in relation to those amendments.
 * ============================================================================
 */

/*
 * Copyright (c) 2001-2004 Swedish Institute of Computer Science.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
 * SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * This file is part of and a contribution to the lwIP TCP/IP stack.
 *
 * Credits go to Adam Dunkels (and the current maintainers) of this software.
 *
 * Christiaan Simons rewrote this file to get a more stable echo example.
 */

#include "telnetd.h"
#include "libtelnet.h"

/* CONSTANTS ***********************************************************************/

#define TELNETD_DEBUG 0

static const telnet_telopt_t my_telopts[] = {
		{ TELNET_TELOPT_ECHO,      TELNET_WONT, TELNET_DONT }, // Echo
		{ TELNET_TELOPT_TTYPE,     TELNET_WILL, TELNET_DONT }, // Terminal Type (tell but don't ask)
		{ TELNET_TELOPT_TSPEED,    TELNET_WONT, TELNET_DONT }, // Terminal Speed (tell and ask)
		{ TELNET_TELOPT_LFLOW,     TELNET_WONT, TELNET_DONT }, // Terminal Speed (tell and ask)
		{ TELNET_TELOPT_BINARY,    TELNET_WILL, TELNET_DO },
		{ -1, 0, 0 }
};

#if LWIP_TCP

static struct tcp_pcb *telnetd_pcb;

/** The server port for telnetd to use */
#ifndef TELNETD_SERVER_PORT
#define TELNETD_SERVER_PORT                   23
#endif

/** The poll delay is X*500ms */
#ifndef TELNETD_POLL_INTERVAL
#define TELNETD_POLL_INTERVAL                 4
#endif

/** Priority for tcp pcbs created by telnetd (very low by default).
 *  Lower priorities get killed first when running out of memroy.
 */
#ifndef TELNETD_TCP_PRIO
#define TELNETD_TCP_PRIO                      TCP_PRIO_MIN
#endif

enum telnetd_states
{
	ES_NONE = 0,
	ES_ACCEPTED,
	ES_RECEIVED,
	ES_CLOSING
};

/* GLOBAL VARIABLES ****************************************************************/

/* LOCAL VARIABLES *****************************************************************/

struct telnetd_state
{
	u8_t state;
	u8_t retries;
	struct tcp_pcb *pcb;
	// Handle for libtelnet.
	telnet_t *libtelnet;
	// Main program handle change callback.
	telnetd_handle_cb handle_cb;
	telnetd_receive_cb receive_cb;
	telnetd_error_cb error_cb;
};

/* MACROS **************************************************************************/

#if TELNETD_DEBUG
#define TELNETD_DEBUG_PRINTF(...) do {tfp_printf(__VA_ARGS__);} while (0)
#else
#define TELNETD_DEBUG_PRINTF(...)
#endif

/* LOCAL FUNCTIONS / INLINES *******************************************************/

// TCP callbacks.
// Forward declarations.
static err_t telnetd_tcp_accept_cb(void *arg, struct tcp_pcb *newpcb, err_t err);
static err_t telnetd_tcp_recv_cb(void *arg, struct tcp_pcb *tpcb, struct pbuf *p, err_t err);
static void telnetd_tcp_error_cb(void *arg, err_t err);
static err_t telnetd_tcp_poll_cb(void *arg, struct tcp_pcb *tpcb);
static err_t telnetd_tcp_sent_cb(void *arg, struct tcp_pcb *tpcb, u16_t len);

static void telnetd_process_recv(struct pbuf *p, struct telnetd_state *telnetdState);
static void telnetd_close(struct tcp_pcb *tpcb, struct telnetd_state *telnetdState);

/** @brief Handle events from the telnet library.
 *  @details Function which handles all events coming from the telnet library.
 */
static void telnetd_event(telnet_t *libtelnet, telnet_event_t *ev, void *user_data)
{
	struct telnetd_state *telnetdState = (struct telnetd_state *)user_data;

	switch (ev->type) {
	case TELNET_EV_DATA:
		// Receive data callback to application.
		if (telnetdState->receive_cb)
		{
			// Note that this receives a pointer to the pbuf data that is
			// obtained from the lwIP library.
			telnetdState->receive_cb((uint8_t *)ev->data.buffer,
					ev->data.size,
					telnetdState);
		}
		break;
	case TELNET_EV_SEND:
		// Send data to remote via TCP.
		tcp_sent(telnetdState->pcb, telnetd_tcp_sent_cb);
		// Not that this receives a buffer obtained from the main application.
		// Hence the copy flag is enabled.
		tcp_write(telnetdState->pcb, ev->data.buffer, (uint16_t)ev->data.size, TCP_WRITE_FLAG_COPY);
		tcp_output(telnetdState->pcb);
		break;
	case TELNET_EV_ERROR:
		// Report error.
		TELNETD_DEBUG_PRINTF("EV_ERROR: %s", ev->error.msg);
		break;
	case TELNET_EV_WILL:
		// Don't care.
		break;
	case TELNET_EV_WONT:
		// Don't care.
		break;
	case TELNET_EV_DO:
		// Don't care.
		break;
	case TELNET_EV_DONT:
		// Don't care.
		break;
	default:
		// Don't care.
		break;
	}
}

/** @brief Telnet connection accepted callback.
 *  @returns Status of connection.
 */
static err_t telnetd_tcp_accept_cb(void *arg, struct tcp_pcb *newpcb, err_t err)
{
	err_t ret_err;
	struct telnetd_state *telnetdState;
	struct telnetd_setup *config;

	config = (struct telnetd_setup *)arg;
	LWIP_UNUSED_ARG(err);

	TELNETD_DEBUG_PRINTF("telnetd_tcp_accept_cb\n");

	tcp_setprio(newpcb, TELNETD_TCP_PRIO);

	telnetdState = (struct telnetd_state *)mem_malloc(sizeof(struct telnetd_state));
	if (telnetdState != NULL)
	{
		telnetdState->state = ES_ACCEPTED;
		telnetdState->pcb = newpcb;
		telnetdState->retries = 0;
		telnetdState->handle_cb = config->hcb;
		telnetdState->receive_cb = config->rcb;
		telnetdState->error_cb = config->ecb;

		// Pass newly allocated state to our callbacks.
		tcp_arg(newpcb, telnetdState);
		tcp_recv(newpcb, telnetd_tcp_recv_cb);
		tcp_err(newpcb, telnetd_tcp_error_cb);
		tcp_poll(newpcb, telnetd_tcp_poll_cb, TELNETD_POLL_INTERVAL);
		ret_err = ERR_OK;

		// Open telnet library and keep handle in telnetdState.
		telnetdState->libtelnet = telnet_init(my_telopts, telnetd_event, 0, telnetdState);

		if (config->hcb)
		{
			config->hcb(telnetdState);
		}
	}
	else
	{
		ret_err = ERR_MEM;
	}
	return ret_err;
}

/** @brief Telnet received data callback
 *  @returns Status of operation.
 */
static err_t telnetd_tcp_recv_cb(void *arg, struct tcp_pcb *tpcb, struct pbuf *p, err_t err)
{
	struct telnetd_state *telnetdState;
	err_t ret_err;

	TELNETD_DEBUG_PRINTF("telnetd_tcp_recv_cb start\n");

	LWIP_ASSERT("arg != NULL",arg != NULL);
	telnetdState = (struct telnetd_state *)arg;

	// A null response is a closed connection.
	if (p == NULL)
	{
		TELNETD_DEBUG_PRINTF("telnetd_tcp_recv_cb p==NULL\n");
		/* remote host closed connection */
		telnetdState->state = ES_CLOSING;

		telnetd_close(tpcb, telnetdState);
		ret_err = ERR_OK;
	}
	else if(err != ERR_OK)
	{
		TELNETD_DEBUG_PRINTF("telnetd_tcp_recv_cb cleanup\n");
		/* cleanup, for unknown reason */
		if (p != NULL)
		{
			TELNETD_DEBUG_PRINTF("pbuf_free %p (%d)\n", p, p->ref);
			pbuf_free(p);
		}
		ret_err = err;
	}
	else if(telnetdState->state == ES_ACCEPTED)
	{
		TELNETD_DEBUG_PRINTF("telnetd_tcp_recv_cb accepted\n");
		/* first data chunk in p->payload */
		telnetdState->state = ES_RECEIVED;

		telnetd_process_recv(p, telnetdState);
		tcp_recved(tpcb, p->tot_len);
		TELNETD_DEBUG_PRINTF("pbuf_free %p (%d)\n", p, p->ref);
		pbuf_free(p);
		ret_err = ERR_OK;
	}
	else if (telnetdState->state == ES_RECEIVED)
	{
		TELNETD_DEBUG_PRINTF("telnetd_tcp_recv_cb received\n");

		telnetd_process_recv(p, telnetdState);
		tcp_recved(tpcb, p->tot_len);
		TELNETD_DEBUG_PRINTF("pbuf_free %p (%d)\n", p, p->ref);
		pbuf_free(p);
		ret_err = ERR_OK;
	}
	else if(telnetdState->state == ES_CLOSING)
	{
		TELNETD_DEBUG_PRINTF("telnetd_tcp_recv_cb closing\n");

		/* odd case, remote side closing twice, trash data */
		tcp_recved(tpcb, p->tot_len);
		TELNETD_DEBUG_PRINTF("pbuf_free %p (%d)\n", p, p->ref);
		pbuf_free(p);
		ret_err = ERR_OK;
	}
	else
	{
		TELNETD_DEBUG_PRINTF("telnetd_tcp_recv_cb unknown trash data\n");

		/* unkown telnetdState->state, trash data  */
		tcp_recved(tpcb, p->tot_len);
		TELNETD_DEBUG_PRINTF("pbuf_free %p (%d)\n", p, p->ref);
		pbuf_free(p);
		ret_err = ERR_OK;
	}

	TELNETD_DEBUG_PRINTF("telnetd_tcp_recv_cb end\n");
	return ret_err;
}

/** @brief Telnet error callback
 */
static void telnetd_tcp_error_cb(void *arg, err_t err)
{
	struct telnetd_state *telnetdState;

	LWIP_UNUSED_ARG(err);

	TELNETD_DEBUG_PRINTF("telnetd_tcp_error_cb\n");

	telnetdState = (struct telnetd_state *)arg;
	if (telnetdState != NULL)
	{
		mem_free(telnetdState);
	}
}

/** @brief Telnet poll callback
 *  @returns Status of operation.
 */
static err_t telnetd_tcp_poll_cb(void *arg, struct tcp_pcb *tpcb)
{
	err_t ret_err;
	struct telnetd_state *telnetdState;

	TELNETD_DEBUG_PRINTF("telnetd_tcp_poll_cb\n");

	telnetdState = (struct telnetd_state *)arg;
	if (telnetdState != NULL)
	{
		/* no remaining pbuf (chain)  */
		if(telnetdState->state == ES_CLOSING)
		{
			telnetd_close(tpcb, telnetdState);
		}
		ret_err = ERR_OK;
	}
	else
	{
		/* nothing to be done */
		tcp_abort(tpcb);
		ret_err = ERR_ABRT;
	}
	return ret_err;
}

/** @brief Telnet data sent callback
 *  @returns Status of operation.
 */
static err_t telnetd_tcp_sent_cb(void *arg, struct tcp_pcb *tpcb, u16_t len)
{
	struct telnetd_state *telnetdState;

	LWIP_UNUSED_ARG(len);

	telnetdState = (struct telnetd_state *)arg;
	telnetdState->retries = 0;

	/* no more pbufs to send */
	if(telnetdState->state == ES_CLOSING)
	{
		telnetd_close(tpcb, telnetdState);
	}

	return ERR_OK;
}

/** @brief Telnet process received packet function
 */
static void telnetd_process_recv(struct pbuf *p, struct telnetd_state *telnetdState)
{
	TELNETD_DEBUG_PRINTF("telnetd_process_recv\n");

	while (p)
	{
		TELNETD_DEBUG_PRINTF("telnetd_process_recv %d bytes at %p\n", p->len, p->payload);
		telnet_recv(telnetdState->libtelnet, p->payload, p->len);
		p = p->next;
	}
}

/** @brief Telnet close function
 */
static void telnetd_close(struct tcp_pcb *tpcb, struct telnetd_state *telnetdState)
{
	TELNETD_DEBUG_PRINTF("telnetd_close!\n");

	telnet_free(telnetdState->libtelnet);
	telnetdState->libtelnet = NULL;

	tcp_arg(tpcb, NULL);
	tcp_sent(tpcb, NULL);
	tcp_recv(tpcb, NULL);
	tcp_err(tpcb, NULL);
	tcp_poll(tpcb, NULL, 0);

	if (telnetdState != NULL)
	{
		mem_free(telnetdState);
	}
	tcp_close(tpcb);
}

/* FUNCTIONS ***********************************************************************/

/** @brief Telnet send callback
 *  @details Called from external code to transmit requested data to the remote
 *  	telnet client.
 */
void telnetd_send(uint8_t *buf, size_t len, void *handle)
{
	struct telnetd_state *telnetdState = (struct telnetd_state *)handle;

	if (handle != NULL)
	{
		TELNETD_DEBUG_PRINTF("telnetd_send %d bytes at %p\n", len, buf);
		telnet_send(telnetdState->libtelnet, (char *)buf, len);
	}
}

/** @brief Query telnet remote client IP address
 *  @returns The IP address of the remote telnet client.
 */
ip_addr_t telnetd_remote_ip(void *handle)
{
	struct telnetd_state *telnetdState = (struct telnetd_state *)handle;
	ip_addr_t erval;


	if (telnetdState->libtelnet)
	{
		if (telnetdState->pcb->remote_ip.addr != 0)
		{
			return telnetdState->pcb->remote_ip;
		}
	}

	erval.addr = 0;
	return erval;
}

/** @brief Telnet initialisation function
 *  @details Initialises the network connection for telnetd server and
 *  	sets up detection of a client connection.
 */
void telnetd_init(struct telnetd_setup *config)
{
	telnetd_pcb = tcp_new();
	if (telnetd_pcb != NULL)
	{
		err_t err;

		tcp_arg(telnetd_pcb, config);
		err = tcp_bind(telnetd_pcb, IP_ADDR_ANY, TELNETD_SERVER_PORT);
		if (err == ERR_OK)
		{
			telnetd_pcb = tcp_listen(telnetd_pcb);
			tcp_accept(telnetd_pcb, telnetd_tcp_accept_cb);

			if (config)
			{
				if (config->hcb)
				{
					config->hcb(NULL);
				}
			}
		}
		else
		{
			/* abort? output diagnostic? */
		}
	}
	else
	{
		/* abort? output diagnostic? */
	}
}

#endif /* LWIP_TCP */
